// mapTopicModel.cpp
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <random>
#include <Eigen/Dense>
#include <Eigen/Sparse>
using namespace Eigen;
using namespace std;
int main(int argc, char* argv[]){
  string  fname = argv[1];            // 処理対象の文書ファイル
  ifstream ifile( fname );
  double  alpha = stod(argv[2]);      // パラメータα
  double   beta = stod(argv[3]);      // パラメータβ
  int      numc = stoi(argv[4]);      // トピック数
  int      seed = stoi(argv[5]);      // 乱数の種  
  string fTheta = argv[6];  // 推定したモデルパラメータθの出力ファイル
  string fPhi   = argv[7];  // 推定したモデルパラメータφの出力ファイル
  int nRows = 0, nCols = 0; // 行数（ndim）と列数（文書数）を表す変数の宣言
  int row, col;
  double val;
  string buf;
  typedef Triplet<double> T;
  vector<T> triplets;
  while( getline(ifile,buf) ){
    istringstream iss(buf);
    iss >> row >> col >> val;
    triplets.emplace_back(T(row-1,col-1,val));
    if( row > nRows ) nRows = row;
    if( col > nCols ) nCols = col;
  }
  ifile.close();
  int ndim = nRows;
  int nvec = nCols;
  SparseMatrix<double> spX(ndim, nvec);
  spX.setFromTriplets(triplets.begin(), triplets.end());
  double B = 0;
  for( int i = 0 ; i < nvec ; i++ )
    for( SparseMatrix<double>::InnerIterator it(spX,i);it; ++it )
      B += it.value();
  SparseMatrix<double> spXt = spX.transpose();
  MatrixXd Ph = MatrixXd::Zero(ndim,numc);  // パラメータφ
  MatrixXd Th = MatrixXd::Zero(numc,nvec);  // パラメータθ
  mt19937 gen(seed);
  uniform_real_distribution<double> dist( 0.001, 1.0 );
  for( int m = 0 ; m < ndim ; m++ )   // φの乱数による初期化 
    for( int k = 0 ; k < numc ; k++ ) Ph(m,k) = dist(gen);
  for( int k = 0 ; k < numc ; k++ )   // θの乱数による初期化 
    for( int i = 0 ; i < nvec ; i++ ) Th(k,i) = dist(gen);
  //-- パラメータ推定
  for( int ic = 0 ; ic < 500 ; ic++ ){   
    MatrixXd Ph1 = (beta - 1.0 ) * MatrixXd::Ones(ndim,numc);
    MatrixXd Th1 = (alpha - 1.0) * MatrixXd::Ones(numc,nvec);
    for( int m = 0 ; m < ndim ; m++ )        // φの更新
      for( SparseMatrix<double>::InnerIterator it(spXt,m);it; ++it ) {
        double sumQ = Ph.row(m).dot(Th.col(it.row()));
        for( int k = 0 ; k < numc ; k++ )
          Ph1(m,k) += it.value() * Ph(m,k)*Th(k,it.row()) / sumQ;
      }
    for( int k = 0 ; k < numc ; k++ )  Ph1.col(k) /= Ph1.col(k).sum();
    for( int i = 0 ; i < nvec ; i++ )        // θの更新
      for( SparseMatrix<double>::InnerIterator it(spX,i);it; ++it ) {
        double sumQ = Ph.row(it.row()).dot(Th.col(i));
        for( int k = 0 ; k < numc ; k++ )
          Th1(k,i) += it.value() * Ph(it.row(),k) * Th(k,i) / sumQ;
      }
    for( int i = 0 ; i < nvec ; i++ ) Th1.col(i) /= Th1.col(i).sum();
    Ph = Ph1;
    Th = Th1;
  }
  double perplexity = 0.0;
  for( int i = 0 ; i < nvec ; i++ )
    for( SparseMatrix<double>::InnerIterator it(spX,i);it; ++it ){
      double xhat = Ph.row(it.row()).dot(Th.col(it.col()));
      perplexity += it.value() * log(xhat);
    }
  perplexity = exp(-perplexity / B);
  cout << "perplexity = " << perplexity << endl;
  ofstream ofileTheta( fTheta );
  ofstream ofilePhi( fPhi );
  ofileTheta << Th << endl;
  ofilePhi   << Ph << endl;  
}
